import bpy
import mathutils
from .components_core import *
from . import preview_meshes
import bmesh
from mathutils import Matrix


def get_field_id(coll_name, field_name):
    return f"PCOL_{coll_name}_{field_name}"

class ColliderField:
    def __init__(self, name, default, subtype = None, icon = "NONE", entity = False, ui_param = None, display_name = None):
        self.name = name
        self.default = default
        self.subtype = subtype
        self.icon = icon
        self.entity = entity
        self.ui_param = ui_param
        self.display_name = display_name
    
    def get_id(self, collname):
        return get_field_id(collname, self.name)
    
    # draws the field in the UI
    def draw(self, obj, collname, container, edit_field):
        r = container.row()
        dname = self.display_name if self.display_name != None else self.name
        if dname != "" or self.icon: r.label(text=dname, icon=self.icon)
        cprop = self.get_id(collname)
        if cprop in obj.keys():
            r.column().prop(obj, '["' + cprop + '"]', text="")
        else:
            r.operator(UpdateCollidersOperator.bl_idname, text="UPDATE", icon="ERROR").name = collname
    
    def get_value(self, obj, collname):
        return obj.get(self.get_id(collname))
    
    def set_value(self, obj, collname, value):
        obj[self.get_id(collname)] = value
    
    # ensures the field exists on the custom data of the object, with the right type
    def update(self, obj, collname):
        cprop = self.get_id(collname)
        if not cprop in obj.keys():
            obj[cprop] = self.default
        if self.entity:
            obj.id_properties_ui(cprop).update(id_type = "OBJECT")
        elif self.subtype != None:
            obj.id_properties_ui(cprop).update(subtype = self.subtype)
        if self.ui_param != None:
            obj.id_properties_ui(cprop).update(**self.ui_param)

class BtnColliderField(ColliderField):
    def __init__(self, name, default, subtype=None, icon="NONE", entity=False, ui_param=None, display_name=None, gizmo_fieldname=None, btn_vert_layout=True):
        super().__init__(name, default, subtype, icon, entity, ui_param, display_name)
        self.gizmo_fieldname = gizmo_fieldname or self.name
        self.btn_vert_layout = btn_vert_layout
    
    def op_btn(self, parent, collname, obj, editing):
        if editing:
            print(collname)
        op = parent.operator(SetGizmosOperator.bl_idname, icon="GIZMO", text="Edit", depress=editing)
        op.collname = collname
        op.fieldname = self.gizmo_fieldname
        op.set_enabled = not editing

    def draw(self, obj, collname, container, edit_field):
        r = container.row()
        dname = self.display_name if self.display_name != None else self.name
        if self.btn_vert_layout:
            lb_parent = r.column()
        else:
            lb_parent = r
        if dname != "" or self.icon: lb_parent.label(text=dname, icon=self.icon)
        self.op_btn(lb_parent, collname, obj, edit_field == self.gizmo_fieldname)
        cprop = self.get_id(collname)
        if cprop in obj.keys():
            r.column().prop(obj, '["' + cprop + '"]', text="")
        else:
            r.operator(UpdateCollidersOperator.bl_idname, text="UPDATE", icon="ERROR").name = collname


class ColliderBase:
    def __init__(self, typename, icon = "NONE"):
        self.typename = typename
        self.icon = icon
        self.translation_enable = ColliderField("pos_enable", False, display_name="Use Offset")
        self.translation = ColliderField("pos", (0.0, 0.0, 0.0), "TRANSLATION", display_name="Offset")
        self.rotation_enable = ColliderField("rot_enable", False, display_name="Use Rotation")
        self.rotation = ColliderField("rot", (0.0, 0.0, 0.0), "EULER", display_name="Rotation")
        self.scale_enable = ColliderField("scale_enable", False, display_name="Use Scale")
        self.scale = ColliderField("scale", (1.0, 1.0, 1.0), display_name="Scale")
        self.fields : list[ColliderField] = [
            self.translation_enable,
            self.translation,
            self.rotation_enable,
            self.rotation,
            self.scale_enable,
            self.scale
        ]
    
    def set_field_map(self):
        self.field_map : dict[str, ColliderField] = {}
        for f in self.fields:
            self.field_map[f.name] = f
    
    def on_edit_start(self, obj, collname, fieldname):
        if fieldname == self.translation.name:
            self.translation_enable.set_value(obj, collname, True)
        elif fieldname == self.rotation.name:
            self.rotation_enable.set_value(obj, collname, True)
        elif fieldname == self.scale.name:
            self.scale_enable.set_value(obj, collname, True)
    
    def remove_all_fields(self, obj, name):
        prefix = get_field_id(name, "")
        for propname in list(obj.keys()):
            if propname.startswith(prefix):
                del obj[propname]

    def remove_unused_fields(self, obj, name):
        prefix = get_field_id(name, "")
        all_fields = {f.name for f in self.fields}
        for propname in list(obj.keys()):
            if propname.startswith(prefix):
                fieldname = propname[len(prefix):]
                if fieldname not in all_fields:
                    del obj[propname]

    def draw_fields(self, obj, name, container):
        edit_coll, edit_field = get_gizmos_editing(obj) or (None, None)
        if edit_coll == name: editing = edit_field
        else: editing = None
        for field in self.fields:
            if (field == self.translation and not self.translation_enable.get_value(obj, name)): continue
            if (field == self.rotation and not self.rotation_enable.get_value(obj, name)): continue
            if (field == self.scale and not self.scale_enable.get_value(obj, name)): continue
            field.draw(obj, name, container, edit_field)
    
    def draw(self, obj, collname, container):
        box = container.box()
        hdr = box.row()
        hdr.label(text=self.typename, icon = self.icon or "NONE")
        hdr.operator(UpdateCollidersOperator.bl_idname, icon="FILE_REFRESH", text="").name = collname
        hdr.operator(DeleteColliderOperator.bl_idname, icon="X", text="").name = collname
        self.draw_fields(obj, collname, box)

    def update_fields(self, obj, name):
        for field in self.fields:
            field.update(obj, name)
        self.remove_unused_fields(obj, name)
    
    def get_enable(self, field : ColliderField, obj, collname):
        if field not in self.fields: return False
        return field.get_value(obj, collname)
    
    def set_gizmo_transform(self, gizmo, obj, collname, offset=None, scale=None):
        def av(a, b): return (ai + bi for ai, bi in zip(a, b))
        def mv(a, b): return (ai * bi for ai, bi in zip(a, b))
        if not offset: offset = (0.0,0.0,0.0)
        if not scale: scale = (1.0,1.0,1.0)
        rotation = (0.0,0.0,0.0)
        if self.get_enable(self.translation_enable, obj, collname):
            offset = av(offset, self.translation.get_value(obj, collname))
        if self.get_enable(self.rotation_enable, obj, collname):
            rotation = self.rotation.get_value(obj, collname)
        if self.get_enable(self.scale_enable, obj, collname):
            scale = mv(self.scale.get_value(obj, collname), scale)
        gizmo.matrix_offset = mathutils.Matrix.LocRotScale(offset, mathutils.Euler(rotation), scale)


    def draw_gizmo(self, gizmo, obj, collname, context):
        pass

collider_types : dict[str, ColliderBase] = {}
def add_collider_type(tp : ColliderBase):
    collider_types[tp.typename] = tp

class SphereCollider(ColliderBase):
    def __init__(self):
        super().__init__("Sphere", icon="SPHERE")
        self.radius = ColliderField("Radius", 1.0)
        self.fields.append(self.radius)
        self.set_field_map()

    def draw_gizmo(self, gizmo, obj, collname, context):
        radius = self.radius.get_value(obj, collname)
        if not radius or radius <= 0: return
        self.set_gizmo_transform(gizmo, obj, collname, scale=(radius, radius, radius))
        gizmo.draw_custom_shape(gizmo.sphere_shape)

add_collider_type(SphereCollider())

class CapsuleCollider(ColliderBase):
    def __init__(self):
        super().__init__("Capsule", icon="MESH_CAPSULE")
        self.radius = ColliderField("Radius", 1.0)
        self.height = ColliderField("Height", 1.0)
        self.fields.append(self.radius)
        self.fields.append(self.height)
        self.set_field_map()
    
    def draw_gizmo(self, gizmo, obj, collname, context):
        radius = self.radius.get_value(obj, collname)
        height = self.height.get_value(obj, collname)
        if not radius or radius <= 0 or not height or height < 0:
            return
        self.set_gizmo_transform(gizmo, obj, collname, (0, 0, height / 2), (radius, radius, radius))
        gizmo.draw_custom_shape(gizmo.capsule_end_shape)
        self.set_gizmo_transform(gizmo, obj, collname, (0, 0, -height / 2), (radius, radius, -radius))
        gizmo.draw_custom_shape(gizmo.capsule_end_shape)
        self.set_gizmo_transform(gizmo, obj, collname, scale=(radius, radius, height/2))
        gizmo.draw_custom_shape(gizmo.capsule_mid_shape)

add_collider_type(CapsuleCollider())

class MeshCollider(ColliderBase):
    def __init__(self):
        super().__init__("Mesh", icon="SPHERE")
        self.convex = ColliderField("Convex", False)
        self.fields.append(self.convex)
        self.set_field_map()

    def create_convex_hull_data(self, obj):
        temp_mesh = obj.to_mesh(preserve_all_data_layers=True, depsgraph=bpy.context.evaluated_depsgraph_get())
        bm = bmesh.new()
        bm.from_mesh(temp_mesh)
        
        # Create the convex hull
        bmesh.ops.convex_hull(bm, input=bm.verts)
        
        # Extract vertices and edges from the convex hull
        vertices = [v.co for v in bm.verts]
        edges = [(e.verts[0].index, e.verts[1].index) for e in bm.edges]
        bm.free()
        return vertices, edges

    def draw_gizmo(self, gizmo, obj, collname, context):

        if self.convex.get_value(obj, collname):
            if obj and obj.type == 'MESH':
                # Generate convex hull vertices and edges
                verts, edges1 = self.create_convex_hull_data(obj)

                if verts and edges1:
                    # Create a list of line segments from the edges
                    lines = []
                    for edge in edges1:
                        lines.append(verts[edge[0]])
                        lines.append(verts[edge[1]])

                    # Create a custom shape for the gizmo
                    self.custom_shape = gizmo.new_custom_shape("LINES", lines)
                
                    # Draw the custom shape
                
                    self.set_gizmo_transform(gizmo, obj, collname)
                    gizmo.draw_custom_shape(self.custom_shape)
        else:
            if obj and obj.type == 'MESH':
                # Generate convex hull vertices and edges
                bm = bmesh.new()
                bm.from_mesh(obj.data)

                # Extract vertices and edges
                vertices = [v.co for v in bm.verts]
                edges = [(e.verts[0].index, e.verts[1].index) for e in bm.edges]

                if vertices and edges:
                    # Create a list of line segments from the edges
                    lines = []
                    for edge in edges:
                        lines.append(vertices[edge[0]])
                        lines.append(vertices[edge[1]])

                    # Create a custom shape for the gizmo
                    self.custom_shape = gizmo.new_custom_shape("LINES", lines)
                
                    # Draw the custom shape
                    self.set_gizmo_transform(gizmo, obj, collname)
                    gizmo.draw_custom_shape(self.custom_shape)


add_collider_type(MeshCollider())

class BoxCollider(ColliderBase):
    def __init__(self):
        super().__init__("Box", icon="CUBE")
        self.halfextents = ColliderField("HalfExtents", (1.0, 1.0, 1.0))
        self.fields.pop()
        self.fields.pop()
        self.fields.append(self.halfextents)
        self.set_field_map()
    
    def draw_gizmo(self, gizmo, obj, collname, context):
        half_extents = self.halfextents.get_value(obj, collname)
        if not half_extents: return
        self.set_gizmo_transform(gizmo, obj, collname, scale=half_extents)
        gizmo.draw_custom_shape(gizmo.box_shape)

class ColliderPreviewWidget(bpy.types.Gizmo):
    bl_idname = "VIEW3D_GT_collider_widget"

    def setup(self):
        self.use_draw_scale = False
        if not hasattr(self, "box_shape"):
            self.box_shape = self.new_custom_shape("LINES", preview_meshes.BOX_MESH_LINES)
        if not hasattr(self, "sphere_shape"):
            self.sphere_shape = self.new_custom_shape("LINES", preview_meshes.SPHERE_MESH_LINES)
        if not hasattr(self, "capsule_end_shape"):
            self.capsule_end_shape = self.new_custom_shape("LINES", preview_meshes.CAPSULE_END_LINES)
        if not hasattr(self, "capsule_mid_shape"):
            self.capsule_mid_shape = self.new_custom_shape("LINES", preview_meshes.CAPSULE_MID_LINES)
    
    def draw(self, context):
        obj = context.active_object
        if not obj: return
        collnames = list_collider_names(obj)
        if not collnames: return
        for collname in collnames:
            type = get_coll_type(collname)
            if type:
                type.draw_gizmo(self, obj, collname, context)

def poll_gizmo(context, tgtfield, coll_type = None):
    obj = context.active_object
    if not obj: return False
    collname, fieldname = get_gizmos_editing(obj)
    if coll_type != None and collname.split("_")[0] != coll_type: return False
    return fieldname == tgtfield

# class ColliderPositionGizmoGroup(bpy.types.GizmoGroup):
#     bl_idname = "VIEW3D_GGT_collider_offset"
#     bl_label = "Collider Offset"
#     bl_space_type = 'VIEW_3D'
#     bl_region_type = 'WINDOW'
#     bl_options = {'3D', 'PERSISTENT'}

#     @classmethod
#     def poll(cls, context):
#         return poll_gizmo(context, "pos")
    
#     def setup(self, context):     
#         self.gx = self.gizmos.new("GIZMO_GT_dial_3d")
#         #self.gy = self.gizmos.new("GIZMO_GT_arrow_3d")
#         #self.gz = self.gizmos.new("GIZMO_GT_blank_3d")

#         self.gx.color = 1,0,0
#         #self.gy.color = 0,1,0
#         #self.gz.color = 0,0,1

#     def refresh(self, context):
#         obj = context.object
#         mat = obj.matrix_world.normalized()
#         obj = context.active_object
#         if not obj: return
#         collname, fieldname = get_gizmos_editing(obj)
#         if not collname: return
#         ct = get_coll_type(collname)
#         if not ct: return
#         self.gx.matrix_basis = mat
#         #self.gy.matrix_basis = mat
#         #self.gz.matrix_basis = mat
#         ct.set_gizmo_transform(self.gx, obj, collname)
#         pos_id = get_field_id(collname, "pos")
#         self.gx.target_set_prop("offset", obj, '["test_offset"]')
#         self.gx.target_set_operator


class ColliderPreviewGizmoGroup(bpy.types.GizmoGroup):
    bl_idname = "VIEW3D_GGT_box_collider"
    bl_label = "Collider Preview"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'WINDOW'
    bl_options = {'3D', 'PERSISTENT'}

    @classmethod
    def poll(cls, context):
        return obj_has_colliders(context.active_object)
    
    def setup(self, context):
        gz = self.gizmos.new(ColliderPreviewWidget.bl_idname)
        gz.color = 0.0, 1.0, 0.0
        gz.alpha = 0.75

        self.preview_gizmo = gz

    def refresh(self, context):
        ob = context.object
        gz = self.preview_gizmo
        gz.matrix_basis = ob.matrix_world.normalized()



add_collider_type(BoxCollider())

coll_type_enum = [
    (name, name, "", i+1)
    for i, name in enumerate(sorted(collider_types.keys()))
]

def list_collider_names(obj):
    st = obj.get("penicillin_colliders", None)
    if not st: return list()
    return list(st.split(","))

def set_collider_names(obj, names):
    if names == None or len(names) == 0:
        if "penicillin_colliders" in obj.keys():
            del obj["penicillin_colliders"]
    else:
        obj["penicillin_colliders"] = ",".join(names)

def add_collider(obj, coll_type):
    names = list_collider_names(obj)
    i = 0
    while f"{coll_type}_{i}" in names:
        i += 1
    new_name = f"{coll_type}_{i}"
    names.append(new_name)
    set_collider_names(obj, names)
    collider_types[coll_type].update_fields(obj, new_name)
    update_colliders_panel()

def get_coll_type(collname):
    return collider_types.get(collname.split("_")[0])

def remove_collider(obj, name):
    names = list_collider_names(obj)
    names.remove(name)
    set_collider_names(obj, names)
    type = collider_types.get(name.split("_")[0])
    if type: type.remove_all_fields(obj, name)
    update_colliders_panel()

def update_collider_fields(obj, name):
    type = name.split("_")[0]
    type = collider_types.get(type)
    if type == None: return
    type.update_fields(obj, name)

def clear_gizmos_editing(obj):
    if "_PCOLL_GIZMO_CUR" in obj.keys():
        del obj["_PCOLL_GIZMO_CUR"]
    update_colliders_panel()

def set_gizmos_editing(obj, collname, fieldname):
    obj["_PCOLL_GIZMO_CUR"] = collname + "|" + fieldname
    tp = get_coll_type(collname)
    if tp:
        tp.on_edit_start(obj, collname, fieldname)
    update_colliders_panel()

def get_gizmos_editing(obj):
    v = obj.get("_PCOLL_GIZMO_CUR")
    if v: return tuple(v.split("|"))
    return (None, None)

def gizmo_type_active(obj, gizmo_type):
    collname, fieldname = get_gizmos_editing()
    if not collname: return False
    type = get_coll_type(collname)
    if not type: return False
    field = type.field_map.get(fieldname)
    if not field: return False

def obj_has_colliders(obj):
    if not obj: return False
    comps = list_components(obj)
    return "Rigidbody" in comps

def update_colliders_panel():
    try:
        bpy.utils.unregister_class(CollidersPanel)
    except:
        pass
    bpy.utils.register_class(CollidersPanel)

class CollidersPanel(bpy.types.Panel):
    bl_idname = "OBJECT_PT_penicillin_colliders"
    bl_label = "Penicillin Colliders"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "object"

    @classmethod
    def poll(cls, context):
        return obj_has_colliders(context.active_object)
    
    def draw(self, context):
        obj = context.active_object
        if not obj: return False
        comps = list_components(obj)
        if "Rigidbody" not in comps: return
        
        self.layout.operator(AddColliderOperator.bl_idname, text=AddColliderOperator.bl_label, icon="ADD")
        names = list_collider_names(obj)
        for name in names:
            type = name.split("_")[0]
            type = collider_types.get(type)
            if type == None: continue
            type.draw(obj, name, self.layout)


class AddColliderOperator(bpy.types.Operator):
    bl_idname = "penicillin.add_collider"
    bl_label = "Add Collider"
    bl_description = "Adds a collider to the object"
    bl_property = "type"

    type: bpy.props.EnumProperty(items=coll_type_enum)

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'RUNNING_MODAL'}

    def execute(self, context):
        add_collider(context.active_object, self.type)
        return {'FINISHED'}

class DeleteColliderOperator(bpy.types.Operator):
    bl_idname = "penicillin.remove_collider"
    bl_label = "Remove Collider"
    bl_icon = "X"

    name: bpy.props.StringProperty()
    
    def execute(self, context):
        remove_collider(context.active_object, self.name)
        return {'FINISHED'}

class UpdateCollidersOperator(bpy.types.Operator):
    bl_idname = "penicllin.update_collider_fields"
    bl_label = "Update Fields"
    bl_description = "Updates the properties to match the defined fields. Can delete data if fields have been removed from definition."
    name: bpy.props.StringProperty()

    def execute(self, context):
        update_collider_fields(context.active_object, self.name)
        return {'FINISHED'}

class SetGizmosOperator(bpy.types.Operator):
    bl_idname = "penicllin.set_gizmos_editing"
    bl_label = "Edit"
    bl_description = "Sets this field as being currently edited by gizmos."
    collname: bpy.props.StringProperty()
    fieldname: bpy.props.StringProperty()
    set_enabled: bpy.props.BoolProperty()

    def execute(self, context):
        if self.set_enabled: set_gizmos_editing(context.active_object, self.collname, self.fieldname)
        else: clear_gizmos_editing(context.active_object)
        return {'FINISHED'}