import bpy, random
from .components_core import *
comptypes = ct.component_types
ctmap = dict((c.name, c) for c in comptypes)

comp_enum_items = [
    (c.name, c.name, "", i+1)
    for i, c in enumerate(comptypes)
]

def update_components(obj, ctype):
    # delete removed fields
    for propname in list(obj.keys()):
        if propname.startswith(ctype.cprop_prefix):
            name_defined = False
            for f in ctype.fields:
                if f.cprop == propname:
                    name_defined = True
                    break
            if not name_defined:
                del obj[propname]
    
    # add any new fields
    for field in ctype.fields:
        if not field.cprop in obj.keys():
            obj[field.cprop] = field.default
        if field.type == "entity":
            obj.id_properties_ui(field.cprop).update(id_type = "OBJECT")
        elif field.subtype != None:
            obj.id_properties_ui(field.cprop).update(subtype = field.subtype)
        if field.ui_param != None:
            obj.id_properties_ui(field.cprop).update(**field.ui_param)

def add_component(obj, typename):
    s = list_components(obj)
    s.add(typename)
    set_components(obj, s)
    update_components(obj, ctmap[typename])
    redraw_panel()

def remove_component(obj, typename):
    s = list_components(obj)
    s.remove(typename)
    set_components(obj, s)

    for f in ctmap[typename].fields:
        if f.cprop in obj.keys():
            del obj[f.cprop]
    
    redraw_panel()

def redraw_panel():
    try:
        bpy.utils.unregister_class(ComponentsPanel)
    except:
        pass
    bpy.utils.register_class(ComponentsPanel)

class ComponentsPanel(bpy.types.Panel):
    bl_idname = "OBJECT_PT_penicillin_components"
    bl_label = "Penicillin Components"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "object"

    def draw(self, context):
        obj = context.active_object
        if obj == None: return 
        self.layout.operator(AddComponentOperator.bl_idname, text=AddComponentOperator.bl_label, icon="ADD")

        comps = list_components(obj)
        for cname in comps:
            ctype = ctmap[cname]
            box = self.layout.box()
            hdr = box.row()
            hdr.label(text=cname, icon=ctype.icon or "OBJECT_DATA")
            hdr.operator(UpdateComponentsOperator.bl_idname, icon="FILE_REFRESH", text="").type = cname
            hdr.operator(DeleteComponentOperator.bl_idname, icon="X", text="").type = cname
            for field in ctype.fields:
                r = box.row()
                r.label(text=field.name, icon=field.icon or "NONE")
                if field.cprop in obj.keys():
                    r.prop(obj, '["' + field.cprop + '"]', text="")
                else:
                    r.operator(UpdateComponentsOperator.bl_idname, text="UPDATE", icon="ERROR").type = cname

class UpdateComponentsOperator(bpy.types.Operator):
    bl_idname = "penicllin.update_component_fields"
    bl_label = "Update Fields"
    bl_description = "Updates the properties to match the defined fields. Can delete data if fields have been removed from definition."
    type: bpy.props.EnumProperty(items=comp_enum_items)

    def execute(self, context):
        update_components(context.active_object, ctmap[self.type])
        return {'FINISHED'}


class DeleteComponentOperator(bpy.types.Operator):
    bl_idname = "penicillin.remove_component"
    bl_label = "Remove Component"
    bl_icon = "X"

    type: bpy.props.EnumProperty(items=comp_enum_items)
    
    def execute(self, context):
        rem_type = ctmap[self.type]
        remove_component(context.active_object, rem_type.name)
        return {'FINISHED'}

class AddComponentOperator(bpy.types.Operator):
    bl_idname = "penicillin.add_component"
    bl_label = "Add Component"
    bl_description = "Adds a penicillin component to the object"
    bl_property = "type"

    type: bpy.props.EnumProperty(items=comp_enum_items)

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'RUNNING_MODAL'}

    def execute(self, context):
        add_type = ctmap[self.type]
        add_component(context.active_object, add_type.name)
        return {'FINISHED'}